module SDWriter # (
    parameter  CLK_DIV = 1, // when clk = 0~25MHz   , set CLK_DIV to 0,
                            // when clk = 25~50MHz  , set CLK_DIV to 1,
                            // when clk = 50~100MHz , set CLK_DIV to 2,
                            // when clk = 100~200MHz, set CLK_DIV to 3,
                            // when clk = 200~400MHz, set CLK_DIV to 4,
                            // ......
    parameter  SIMULATION = 0
) (
    // clock
    input  logic        clk,
    // rst_n active-low
    input  logic        rst_n,
    // SDcard signals (connect to SDcard)
    output logic        sdclk,
    output logic [ 3:0] sddat,
    output logic        sdcmdin,
    output logic        sdcmdout,
    output logic        sdcmdoe,
    // user write sector command interface
    input  logic        wstart,         // pulse high to start write op
    input  logic [31:0] wsector_no,     // user-supplied sector address
    output logic        wbusy,
    output logic        wdone,
    // sector data input interface
    // bytes will be read from DDR3 by SDInterface, with address set by secaddr here
    // plus the usual buffer_addr offset in SDInterface.
    //output logic        bytereq,        // pulse high to request next byte
    output logic [ 3:0] bytePtr,        // points to byte in cache to be written
    input  logic [ 7:0] wbyte,
    //input  logic        byteAck,        // valid data on secByte
    // these signals are direction controls specific to the DECA board
    output logic        SD_CMD_DIR,     // HIGH = TO SD card, LOW = FROM SD card
    output logic        SD_D0_DIR,      // HIGH = TO SD card, LOW = FROM SD card
    output logic        SD_D123_DIR,    // HIGH = TO SD card, LOW = FROM SD card
    output logic        SD_SEL
);

localparam  SLOWCLKDIV = SIMULATION ? 16'd1   : ( (16'd1<<CLK_DIV)*16'd35 ) ;
localparam  FASTCLKDIV = SIMULATION ? 16'd0   : ( (16'd1<<CLK_DIV) )        ;
localparam  CMDTIMEOUT = SIMULATION ? 15'd100 : 15'd500   ; // according to SD datasheet, Ncr(max) = 64 clock cycles, so 500 cycles is enough
localparam  DATTIMEOUT = SIMULATION ? 'd200   : 'd1000000 ; // according to SD datasheet, 1ms is enough to wait for DAT result, here, we set timeout to 1000000 clock cycles = 80ms (when SDCLK=12.5MHz)

// DECA SD direction controls - HIGH for writes to SD card, LOW for reads from SD card
assign  SD_CMD_DIR  = sdcmdoe ;
assign  SD_D0_DIR   = 1'b0    ; // read-only to prevent writes to SD card whilst testing
assign  SD_D123_DIR = 1'b0    ; // read-only to prevent writes to SD card whilst testing

// SD interface voltage - DECA-specific ******* THIS CAN BE MODIFIED SO 1.8V CARDS ARE SUPPORTED ********
assign  SD_SEL      = 1'b0 ; // LOW = 3.3V, HIGH = 1.8V

wire            sddat0      = sddat[0]         ; // only use 1bit mode of SDDAT
wire    [  2:0] rlsb        = 3'd7 - widx[2:0] ;

logic           bytereq              ;
logic           start       = 1'b0   ;
logic           writing     = 1'b0   ;
logic   [  5:0] cmd         = '0     ;
logic   [  7:0] writebyte /* synthesis keep */; // this is the byte to be written to the SD card
logic   [ 15:0] clkdiv      = 16'd50 ;
logic   [ 15:0] precycles   = '0     ;
logic   [ 15:0] rca         = '0     ;
logic   [ 31:0] wsectoraddr = '0     ;
logic   [ 31:0] arg         = '0     ;
logic   [ 31:0] widx        = 0      ;
logic   [ 31:0] resparg              ;
logic   [127:0] resparg_long         ;
logic           busy, done, timeout  ;
logic           syntaxerr            ;
logic           sdclkl      = 1'b0   ;

// TEST VALUES
logic        enable     ;
logic [ 8:0] byte_count ; // maximum value 512
logic [ 7:0] testByte   ;
logic [24:0] timer      ;

//    |---------- CARD INITIALISATION ---------|-SEL-|BLKLEN|
enum { CMD0, CMD8, CMD55_41, ACMD41, CMD2, CMD3, CMD7, CMD16, IDLE, WRITING, WRITING2 } sdstate = IDLE ;
//
enum { UNKNOWN, SDv1, SDv2, SDHCv2, SDv1Maybe } cardtype = UNKNOWN ;
enum { WWAIT, WDURING, WTAIL, WDONE, WTIMEOUT } wstate   = WWAIT   ;

assign bytePtr   = widx[6:3]       ;
assign wbusy     = (sdstate!=IDLE) ;
assign card_stat = sdstate[3:0]    ;
assign card_type = cardtype[1:0]   ;

// Instantiate an SDCmdCtrl instance, using implicit named port connections,
// which are automatically connected to wires/ports of same name with equivalent data types.
SDCmdCtrl #(
    CMDTIMEOUT
) sd_cmd_ctrl_inst (
    .clk,
    .rst_n,
    // user input signal
    .start,
    .precycles,
    .clkdiv,
    .cmd,
    .arg,
    // user output signal
    .resparg,
    .resparg_long,
    .busy,
    .done,
    .timeout,
    .syntaxerr,
    // SD CLK output
    .sdclk,
    // 1bit SD CMD
    .sdcmdoe,
    .sdcmdout,  // <- from cmd_ctrl
    .sdcmdin    // CMD -> cmd_ctrl
);

task automatic set_cmd(input _start, input[15:0] _precycles='0, input[15:0] _clkdiv=SLOWCLKDIV, input[5:0] _cmd='0, input[31:0] _arg='0 ) ;
    start     = _start     ;
    precycles = _precycles ;
    clkdiv    = _clkdiv    ;
    cmd       = _cmd       ;
    arg       = _arg       ;
endtask

always @(posedge clk or negedge rst_n) begin

    if ( ~rst_n ) begin
        byte_count <= 0    ;
        enable     <= 0    ;
        bytereq    <= 0    ;
        timer      <= 0    ;
        wdone      <= 0    ;
        widx       <= 0    ;
    end
    else if ( wstart ) begin
        byte_count <= 0    ;
        enable     <= 1'b1 ;
        bytereq    <= 0    ;
        timer      <= 0    ;
        wdone      <= 0    ;
        widx       <= 0    ;
        sdstate    <= WRITING ;
    end
    else if ( enable ) begin
        timer <= timer + 1'b1 ;
        if (timer > 50) begin
            if ( byte_count < 512 ) begin // request byte
                if ( byte_count == 511 ) begin
                    wdone      <= 1'b1 ;
                    enable     <= 1'b0 ;
                    widx       <= 0    ;
                    sdstate    <= IDLE ;
                end
                else begin
                    byte_count <= byte_count + 1 ;
                    widx       <= widx + 8       ;
                    writebyte  <= wbyte          ;
                end
            end
            timer <= 0 ; // reset timer
        end
    end

end

/*
always @(posedge clk or negedge rst_n) begin

    if ( ~rst_n ) begin
        wdone       = 1'b0    ;
        writing     = 1'b0    ;
        wsectoraddr ='0       ;
        sdstate     = IDLE    ; // SDWriter shouldn't try to access SD card at reset
        cardtype    = UNKNOWN ;
        rca         = '0      ;
    end else if ( wstart ) begin
        writing     = 1'b1    ; // start writing a sector
    end else if ( writing ) begin
        set_cmd(0)   ;
        wdone = 1'b0 ;
        if ( sdstate == WRITING2 ) begin // Handle write timeouts and end process
            if ( wstate == WTIMEOUT ) begin
                set_cmd(1, 16 , FASTCLKDIV, 17, wsectoraddr) ;
                sdstate = WRITING ;
            end
            else if ( wstate == WDONE ) begin
                wdone   = 1'b1 ;
                sdstate = IDLE ;
                writing = 1'b0 ;
            end
        end else if ( busy ) begin // Handle SD card initialisation
            if ( done ) begin
                case ( sdstate )
                CMD0    :   sdstate = CMD8 ;
                CMD8    :   if ( timeout ) begin
                                cardtype = SDv1Maybe ;
                                sdstate  = CMD55_41  ;
                            end else if ( ~syntaxerr && resparg[7:0] == 8'hAA ) begin
                                sdstate  = CMD55_41  ;
                            end
                CMD55_41:   if ( ~timeout && ~syntaxerr ) sdstate = ACMD41 ;
                ACMD41  :   if( ~timeout && ~syntaxerr && resparg[31] ) begin
                                cardtype = (cardtype==SDv1Maybe) ? SDv1 : (resparg[30] ? SDHCv2 : SDv2) ;
                                sdstate  = CMD2      ;
                            end else begin
                                sdstate  = CMD55_41  ;
                            end
                CMD2    : if ( ~timeout && ~syntaxerr ) sdstate = CMD3;
                CMD3    : if ( ~timeout && ~syntaxerr ) begin
                              rca     = resparg[31:16] ;
                              sdstate = CMD7           ;
                          end
                CMD7    : if ( ~timeout && ~syntaxerr ) sdstate = CMD16;
                CMD16   : if ( ~timeout && ~syntaxerr ) sdstate = IDLE;
                WRITING :   if ( ~timeout && ~syntaxerr ) begin
                                sdstate = WRITING2     ;
                            end
                            else begin // THIS IS THE READ COMMAND!!
                                set_cmd(1, 128 , FASTCLKDIV, 17, wsectoraddr);
                            end
                endcase
            end
        end else begin
                case ( sdstate )
                CMD0    :   set_cmd( 1, SIMULATION?16:99999, SLOWCLKDIV,  0, 'h00000000);
                CMD8    :   set_cmd( 1, 20    , SLOWCLKDIV,  8, 'h000001aa  ) ;
                CMD55_41:   set_cmd( 1, 20    , SLOWCLKDIV, 55, 'h00000000  ) ;
                ACMD41  :   set_cmd( 1, 20    , SLOWCLKDIV, 41, 'hc0100000  ) ;
                CMD2    :   set_cmd( 1, 20    , SLOWCLKDIV,  2, 'h00000000  ) ;
                CMD3    :   set_cmd( 1, 20    , SLOWCLKDIV,  3, 'h00000000  ) ;
                CMD7    :   set_cmd( 1, 20    , SLOWCLKDIV,  7, {rca,16'h0} ) ;
                CMD16   :   set_cmd( 1, SIMULATION?20:99999 , FASTCLKDIV, 16, 'h00000200);
                IDLE    :   if ( wstart & ~wbusy ) begin 
                                wsectoraddr = ( cardtype == SDHCv2 ) ? wsector_no : ( wsector_no*512 );
                                // READ COMMAND!!
                                set_cmd ( 1, 32 , FASTCLKDIV, 17, wsectoraddr ) ;
                                sdstate = WRITING ;
                            end
                endcase
        end
    end
end

always @( posedge clk or negedge rst_n ) begin

    if ( ~rst_n ) begin
        sdclkl  <= 1'b0 ;
        wstate  = WWAIT ;
        bytereq  = 1'b0 ;
        widx    = '0    ;
        outaddr = '0    ;
    end else begin
        sdclkl <= sdclk ;
        bytereq  = 1'b0 ;
        outaddr = '0    ;
        if ( sdstate != WRITING && sdstate != WRITING2 ) begin
            wstate = WWAIT ;
            widx   = 0     ;
        end else if ( ~sdclkl & sdclk ) begin  // on rising edge of sdclk
            case ( wstate )
                WWAIT:  begin
                    // Must pull sddat0 LOW to indicate start of write
                    if ( ~sddat0 ) begin
                        wstate = WDURING  ;
                        widx   = 0        ;
                    end else if ( ( ++widx ) > DATTIMEOUT ) begin
                        wstate = WTIMEOUT ;
                    end
                end
                WDURING: begin
                    //sddat0 = outbyte[rlsb] ;  // write byte
                    if ( rlsb == 3'd0 ) begin
                        bytereq  = 1'b1      ; // byte written, get next one
                        // widx[2:0] is the bit pointer
                        // widx[6:3] can be used as dat_ptr in SDInterface
                        outaddr = widx[11:3] ;
                    end
                    if ( ( ++widx ) >= 512*8 ) begin
                        wstate = WTAIL ;
                        widx   = 0     ;
                    end
                end
                WTAIL:  begin
                    if ( ( ++widx ) >= 8*8 ) begin
                        wstate = WDONE ;
                    end
                end
            endcase

        end

    end

end */

endmodule
